# -*- coding: utf-8 -*-
"""
This module defines the classes that represent the types read in an XML
Schema.

The types are defined as a specific node derived from
:class:`~pyxst.xml_struct.xsd.utils.XsdNode`.

.. autodata:: EMPTY

.. autodata:: LIST

.. autodata:: UNION

.. autodata:: ELEMENTS_ONLY

.. autodata:: EXTENSION

.. autodata:: CONSTRAINT_MAPPING

.. autoclass:: XsdType
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: XsdSimpleType
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: XsdSimpleListType
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: XsdSimpleUnionType
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: XsdComplexType
   :members:
   :undoc-members:
   :show-inheritance: 
"""
__docformat__ = "restructuredtext en"


from .errors import XsdError
from .utils import split_qname, MIXED
from .utils import (
    XsdNode, XsdUsage, TYPE, ATTRIBUTE, ATTRIBUTE_GROUP, XSD_NS)
from .groups import XsdSequence
from .constraints import (
    XsdStrictMinConstraint, XsdMinConstraint, XsdMaxConstraint,
    XsdStrictMaxConstraint, XsdMaxDigitsConstraint, XsdMinDigitsConstraint,
    XsdLengthConstraint, XsdMinLengthConstraint, XsdMaxLengthConstraint,
    XsdPossibleValuesConstraint, XsdWhiteSpacePolicyConstraint,
    XsdValuePatternConstraint)


EMPTY = "Empty"
"""
Constant used to define the type of the content that can occur inside a type.
The type doesn't have any content.

Type: :class:`str`
"""

LIST = "List of Textual Items"
"""
Constant used to define the type of the content that can occur inside a type.
The type contains a space separated list of textual items.

Type: :class:`str`
"""

UNION = "Alternative between Textual Types"
"""
Constant used to define the type of the content that can occur inside a type.
The type of the content must be chosen among several textual types.

Type: :class:`str`
"""

ELEMENTS_ONLY = "Only Elements"
"""
Constant used to define the type of the content that can occur inside a type.
The type only contains sub-elements.

Type: :class:`str`
"""

RESTRICTION = "Restriction"
"""
Constant used to define the derivation method between a type and its base
type. The type restricts its base type (adds new constraints).

Type: :class:`str`
"""

EXTENSION = "Extension"
"""
Constant used to define the derivation method between a type and its base
type. The type extends its base type (adds new attributes or sub-elements).

Type: :class:`str`
"""

CONSTRAINT_MAPPING = {u"minExclusive"  : XsdStrictMinConstraint,
                      u"minInclusive"  : XsdMinConstraint,
                      u"maxExclusive"  : XsdStrictMaxConstraint,
                      u"maxInclusive"  : XsdMaxConstraint,
                      u"totalDigits"   : XsdMaxDigitsConstraint,
                      u"fractionDigits": XsdMinDigitsConstraint,
                      u"length"        : XsdLengthConstraint,
                      u"minLength"     : XsdMinLengthConstraint,
                      u"maxLength"     : XsdMaxLengthConstraint,
                      u"enumeration"   : XsdPossibleValuesConstraint,
                      u"whiteSpace"    : XsdWhiteSpacePolicyConstraint,
                      u"pattern"       : XsdValuePatternConstraint,
                      }
"""
Dictionary mapping the name of the constraints found in the simple type
restrictions and the classes that represent these constraints.

Type: dictionary of :class:`class` indexed with :class:`unicode`
"""


class XsdType(XsdNode):
    """
    Abstract class representing a type defined in an XML Schema.

    Two different types can be defined: simple types that only contain textual
    data or complex types that can contain attributes or elements. In an XML
    Schema, it is also possible to define types that are derived from other 
    types (they use the definition of the other type and add or remove
    some content or some constraints).

    After all the types have been read, it is possible to resolve the 
    derivation hierarchy ie propagate in each type the information stored
    in the types it is derived from (see ``resolve_derivation`` method).

    If a type is directly defined inside an element, a name is nevertheless 
    automatically created for this type.

    .. attribute:: content_type

       Type of the content that can be found in the type. If the type is a 
       simple type or a complex type with a simple content, the content is 
       purely textual and therefore this attribute will give the built-in type 
       of this content (string, float, date, etc.)

       If the type is a simple list type or a simple union type, this attribute
       is left emptycomplex type with
       complex content, the content can either be ``ELEMENTS_ONLY`` or
       ``MIXED`` (a mixed content is a content that mixes textual data
       and child elements).

    .. attribute:: derivation_resolved

       Flag set to true after the derivation of this type has been resolved.
       When the derivation is resolved, all the information inherited from the
       base type has been propagated in this type.

    .. attribute:: constraints

       Constraints applied to the base type this type is
       derived from.  Different sorts of constraints can be
       defined. They are stored in a dictionary whose key is the name
       of the constraint (name of the XML constraint element in the
       XML Schema). The possible names of constraints are:
       ``minExclusive``,``minInclusive``, ``maxExclusive``,
       ``maxInclusive``, ``totalDigits``, ``fractionDigits``,
       ``length``, ``minLength``, ``maxLength``, ``enumeration``,
       ``whiteSpace``, ``pattern``.

       Only types whose content is purely textual (simple types or complex
       types with simple content) can have constraints.

       Type: Dictionary of 
       :class:`~pyxst.xml_struct.xsd.constraints.XsdConstraint` 
       indexed with :class:`unicode`

    .. attribute:: _base_type

       Base type that this type is derived from. The derivation method
       is defined with ``_derivation_method`` attribute.

       This attribute is private as it is not necessary to read the information
       in the base type after the derivation of this type has been resolved.

       Type: :class:`~pyxst.xml_struct.xsd.types.XsdType`

    .. attribute:: _deriv_method

       Method of derivation used to define this type from its base type 
       declared in ``_base_type`` attribute. The possible values for this
       attribute are ``RESTRICTION`` and ``EXTENSION``.

       This attribute is private as it is not necessary to read the information
       in the base type after the derivation of this type has been resolved.

       Type: :class:`str`

    For this class, the ``category`` class attribute is set to ``TYPE``.

    .. automethod:: __init__
    """
    category = TYPE

    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new type node.

        :param name:
            Name of the type.
        :type name:
            :class:`unicode`
        """
        super(XsdType, self).__init__(name, uid_prefix=uid_prefix)
        self.content_type = EMPTY
        self.derivation_resolved = False
        self.constraints = {}
        self._base_type = None
        self._deriv_method = None
        if self.__class__ is XsdType and \
                name not in [u"{%s}anyType" % XSD_NS,
                             u"{%s}anySimpleType" % XSD_NS]:
            raise XsdError("Can't use this base type for instantiating %s type"
                           " (reserved to basis types)" % name)

    def add_constraint(self, name, value):
        """
        Adds a constraint whose type is ``name`` and value is ``value`` on the
        type.

        The XML Schema defines a set of possible constraints that can be
        set on the textual content found in a type.

        :param name:
            Name of the constraint. The name corresponds to the name
            of the constraint elements that exist in the XML Schema :
            ``minExclusive``,``minInclusive``, ``maxExclusive``,
            ``maxInclusive``, ``totalDigits``, ``fractionDigits``,
            ``length``, ``minLength``, ``maxLength``, ``enumeration``,
            ``whiteSpace``, ``pattern``.
        :type name:
            :class:`unicode`
        :param value:
            Value defined in the constraint.
        :type name:
            :class:`unicode`
        """
        if value is not None:
            cnstr = self.constraints.setdefault(name,
                                                CONSTRAINT_MAPPING[name]())
            cnstr.define_value(value)

    def resolve_derivation(self):
        """
        Resolves the derivation of this type.

        As the types that are instances of ``XsdType`` are the most elementary
        built-in types that are at the basis of the types hierarchy, we have
        nothing to do.

        This method defines the ``content_type`` attribute.

        As this method must not be called more than once, a flag
        (``derivation_resolved`` attribute) is set after the method
        has been completed. The method is not executed if the flag is
        set.
        """
        if self.derivation_resolved:
            return
        # Type is a basic built-in type (anyType ou anySimpleType)
        ns, local_name = split_qname(self.name)
        self.content_type = local_name
        self.derivation_resolved = True

    @property
    def uid(self):
        return u"{0}/${1}".format(self.uid_prefix, self.name)


class XsdSimpleType(XsdType):
    """
    Class representing a simple type defined in an XML Schema.

    Simple types are types that only contain textual data (no attribute nor
    sub-elements).

    .. automethod:: __init__
    """
    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new simple type node.

        :param name:
            Name of the type.
        :type name:
            :class:`unicode`
        """
        super(XsdSimpleType, self).__init__(name, uid_prefix=uid_prefix)

    def resolve_derivation(self):
        """
        Resolves the derivation of this type by propagating the information
        defined on the base type into this type.

        This method defines the ``content_type`` attribute (even if there is
        no derivation from a base type), propagates the constraints defined
        on the base type into this type.

        As this method must not be called more than once, a flag
        (``derivation_resolved`` attribute) is set after the method
        has been completed. The method is not executed if the flag is
        set.
        """
        if self.derivation_resolved:
            return
        # Resolution of derivation on base type that we derive from
        if self._base_type is not None:
            self._base_type.resolve_derivation()
        # Content type definition
        if self._base_type is not None:
            self.content_type = self._base_type.content_type
        else:
            # Simple type is a built-in type
            ns, local_name = split_qname(self.name)
            self.content_type = local_name
        # Constraints propagation
        if self._base_type is not None:
            for name, cnst in self._base_type.constraints.items():
                if name == u"enumeration":
                    for val in cnst.value:
                        self.add_constraint(u"enumeration", val)
                else:
                    if name not in self.constraints:
                        self.constraints[name] = cnst
        self.derivation_resolved = True


class XsdSimpleListType(XsdSimpleType):
    """
    Class representing a simple type defined in an XML Schema that contains a
    list of items whose type is also a simple one.

    Simple types are types that only contain textual data (no attribute nor
    sub-elements).

    .. attribute:: listitem_type

       Type of the list items contained in the list

       Type: :class:`~pyxst.xml_struct.xsd.types.SimpleType`

    .. automethod:: __init__
    """
    def __init__(self, name, listitem_type=None, uid_prefix=u""):
        """Initializes a new simple list type node.

        :param name: Name of the type.  :type name: :class:`unicode`
            :param listitem_type: Type of the list items contained in
            this list.  :type listitem_type:
            :class:`~pyxst.xml_struct.xsd.types.XsdType`

        """
        super(XsdSimpleListType, self).__init__(name, uid_prefix=uid_prefix)
        self.content_type = LIST
        if not listitem_type is None and not listitem_type.category is TYPE:
            raise XsdError("%s can't contain %s list items (should be another "
                           "type)." % (self, listitem_type))
        self.listitem_type = listitem_type

    def add_constraint(self, name, value):
        """
        Raises an exception as it is impossible to directly set a constraint
        on a list type.
        """
        raise XsdError("Can't set a constraint on %s" % self)

    def resolve_derivation(self):
        """
        Resolves the derivation of this type by propagating the resolution
        onto the list item type.

        This method redefines the ``content_type`` attribute that is always
        set to ``LIST`` and doesn't propagate the constraints as there is
        no constraint on a list type.

        As this method must not be called more than once, a flag
        (``derivation_resolved`` attribute) is set after the method
        has been completed. The method is not executed if the flag is
        set.
        """
        if self.derivation_resolved:
            return
        if self._base_type is not None:
            raise XsdError("%s can't derive from a base type" % self)
        if self.listitem_type is None:
            raise XsdError("%s should contain a list item type" % self)
        self.content_type = LIST
        # Resolution of derivation on the list item type
        self.listitem_type.resolve_derivation()
        self.derivation_resolved = True


class XsdSimpleUnionType(XsdSimpleType):
    """
    Class representing a simple type defined in an XML Schema that is
    a choice between several possible types, each one being a simple type.

    Simple types are types that only contain textual data (no attribute nor
    sub-elements).

    .. attribute:: possible_types

       Set of the possible types that can be chosen.

       Type: Set of :class:`~pyxst.xml_struct.xsd.types.SimpleType`

    .. automethod:: __init__
    """
    def __init__(self, name, possible_types=None, uid_prefix=u""):
        """
        Initializes a new simple union type node.

        :param name:
            Name of the type.
        :type name:
            :class:`unicode`
        :param possible_types:
            Possible types of this union.
        :type possible_types:
            List of :class:`~pyxst.xml_struct.xsd.types.XsdType`
        """
        super(XsdSimpleUnionType, self).__init__(name, uid_prefix=uid_prefix)
        self.content_type = UNION
        self.possible_types = set([])
        if possible_types is not None:
            for typ in possible_types:
                if not typ.category is TYPE:
                    raise XsdError("%s can't contain %s possible type (should "
                                   "be another type)." % (self, typ))
                self.possible_types.add(typ)

    def add_constraint(self, name, value):
        """
        Raises an exception as it is impossible to directly set a constraint
        on an union type.
        """
        raise XsdError("Can't set a constraint on %s" % self)

    def resolve_derivation(self):
        """
        Resolves the derivation of this type by propagating the resolution
        onto each of the possible types.

        This method redefines the ``content_type`` attribute that is always
        set to ``UNION`` and doesn't propagate the constraints as there is
        no constraint on a union type.

        As this method must not be called more than once, a flag
        (``derivation_resolved`` attribute) is set after the method
        has been completed. The method is not executed if the flag is
        set.
        """
        if self.derivation_resolved:
            return
        if self._base_type is not None:
            raise XsdError("%s can't derive from a base type" % self)
        if len(self.possible_types) == 0:
            raise XsdError("%s should contain several possible types" % self)
        self.content_type = UNION
        # Resolution of derivation on all the possible types
        for typ in self.possible_types:
            typ.resolve_derivation()
        self.derivation_resolved = True


class XsdComplexType(XsdType):
    """
    Class representing a complex type defined in an XML Schema.

    Complex types are types that can contain attributes. Complex types
    with complex content can contain sub-elements whereas complex
    types with simple content only contain text data.

    Some complex types with complex content can contain elements and
    textual data. This content is called mixed. Mixed content is
    specified in the ``content_type`` attribute.

    .. attribute:: sub_content

       When the complex type has a complex content, this attribute
       contains an usage node pointing towards a group (sequence, choice or 
       all) that contains the sub-elements and sub-groups that can occur 
       inside the type.

       When the complex type has a simple content, this attribute remains
       empty.

       Type: :class:`~pyxst.xml_struct.xsd.utils.XsdUsage`

    .. attribute:: attributes 

       List of all the attributes of this complex type.

       The uniqueness of each attribute is checked thanks to the ``attr_ids``
       attribute of this class.

       Type: list of 
       :class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`

    .. attribute:: attr_ids

       Set containing the identifiers of the attributes of this complex type. 
       Hence, this complex type can't have two attributes with the same name.

       This attribute is only used to check that new attributes added
       to this complex type don't collide with already defined
       attributes with the same name.

       Type: set of :class:`str`

    .. attribute:: attr_groups

       List of attribute groups used in this complex type.

       As it is much more interesting to directly have all the
       attributes defined in this complex type (instead of having a
       hierarchy of attribute groups), a method
       (``resolve_attribute_groups``) is provided to add all the
       attributes of the attributes groups in the complex type.

       Type: list of 
       :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup`

    .. attribute:: _attr_groups_resolved

       Flag that is true id the attribute groups of this complex type have
       already been resolved (ie if the attributes defined in the attribute
       groups have been directly added to this complex type).

       See ``resolve_attribute_groups`` method.

       Type: :class:`boolean`

    .. automethod:: __init__
    """
    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new complex type node.

        :param name:
            Name of the type.
        :type name:
            :class:`unicode`
        """
        super(XsdComplexType, self).__init__(name, uid_prefix=uid_prefix)
        self.sub_content = None
        self.attributes = []
        self.attr_ids = set()
        self.attr_groups = []
        self._attr_groups_resolved = False

    def add_attribute(self, attr, min_occur=1, max_occur=1):
        """
        Adds the ``attr`` attribute to the complex type.

        This method checks that ``attr`` category is ``ATTRIBUTE``, that
        max_occur is never greater than 1 and that there is not already an
        attribute with the same identifier in this complex type.

        :param attr:
            Attribute node to be added in this complex type.
        :type attr:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttribute`
        :param min_occur:
            Minimum occurence for the attribute (0 or 1)
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence for the attribute (0 or 1)
        :type max_occur:
            :class:`int`
        """
        attr_id = attr.uid
        if not attr.category is ATTRIBUTE:
            raise XsdError("Can't add %s to %s as attribute" % (attr, self))
        if attr_id in self.attr_ids:
            raise XsdError("%s already declared in %s" % (attr, self))
        if max_occur > 1:
            raise XsdError("%s can't occur more than one time in %s"
                           % (attr, self))
        self.attr_ids.add(attr_id)
        self.attributes.append(XsdUsage(attr, min_occur, max_occur))

    def add_attribute_group(self, attr_grp):
        """
        Adds the ``attr_grp`` attribute group to this complex type.

        This method checks ``attr_grp`` category is ``ATTRIBUTE_GROUP``.

        :param attr_grp:
            Attribute group node to be added in this complex type.
        :type attr_grp:
            :class:`~pyxst.xml_struct.xsd.attributes.XsdAttributeGroup` 
        """
        if not attr_grp.category is ATTRIBUTE_GROUP:
            raise XsdError("Can't add %s to %s as attribute group"
                           % (attr_grp, self))
        self.attr_groups.append(attr_grp)

    def resolve_attribute_groups(self):
        """
        Resolves the attribute group references by reading their
        content and adding all the attributes defined in the attribute
        groups into this complex type thanks to the ``add_attribute``
        method.

        As this method must not be called more than once (it is an error to
        define several times an attribute with the same name), a flag 
        (``_attr_groups_resolved`` attribute) is set after the method has been
        completed. The method is not executed if the flag is set.
        """
        if self._attr_groups_resolved:
            return
        for attr_grp in self.attr_groups:
            for attr, min_occur, max_occur in iter(attr_grp):
                self.add_attribute(attr, min_occur, max_occur)
        self._attr_groups_resolved = True

    def resolve_derivation(self):
        """
        Resolves the derivation of this type by propagating the information
        defined on the base type into this type. The propagation depends on
        the derivation method (restriction or extension).

        This method defines the ``content_type`` attribute (even if there is
        no derivation from a base type), propagates the constraints defined
        on the base type into this type, aggregates the inner content defined
        in the ``sub_content`` attribute with the inner content defined in the 
        base type (only if the derivation method is extension) and propagates 
        the attributes defined in the base type into this type.

        As this method must not be called more than once, a flag
        (``derivation_resolved`` attribute) is set after the method
        has been completed. The method is not executed if the flag is
        set.
        """
        if self.derivation_resolved:
            return
        # Resolution of derivation on base type that we derive from
        if self._base_type is not None:
            self._base_type.resolve_derivation()
        # Content type definition
        if self.content_type != MIXED:
            if self._base_type is not None:
                self.content_type = self._base_type.content_type
                if self.sub_content is not None and \
                        len(self.sub_content.referenced_node) > 0:
                    self.content_type = ELEMENTS_ONLY
                elif self._deriv_method == RESTRICTION and \
                        self._base_type.content_type in [MIXED, ELEMENTS_ONLY]:
                    self.content_type = EMPTY
                elif self._deriv_method == EXTENSION and \
                        self._base_type.content_type == MIXED and \
                        self._base_type.sub_content is not None and \
                        len(self._base_type.sub_content.referenced_node) > 0:
                    self.content_type = ELEMENTS_ONLY
            else:
                if self.sub_content is not None and \
                        len(self.sub_content.referenced_node) > 0:
                    self.content_type = ELEMENTS_ONLY
                else:
                    self.content_type = EMPTY
        elif self.sub_content is None:
            # A Mixed content with no sub-element is just a textual content
            self.content_type = u"anySimpleType"
        # Constraints propagation
        if self._base_type is not None:
            for name, cnst in self._base_type.constraints.items():
                if name == u"enumeration":
                    for val in cnst.value:
                        self.add_constraint(u"enumeration", val)
                else:
                    if name not in self.constraints:
                        self.constraints[name] = cnst
        # Sub-content aggregation
        if isinstance(self._base_type, XsdComplexType) and \
                self._base_type.sub_content is not None:
            if self._deriv_method == EXTENSION:
                if self.sub_content is None or \
                        len(self.sub_content.referenced_node) == 0:
                    self.sub_content = self._base_type.sub_content
                else:
                    aggr = XsdSequence("%s/extension" % self.name)
                    aggr.children.append(self._base_type.sub_content)
                    aggr.children.append(self.sub_content)
                    self.sub_content = XsdUsage(aggr, 1, 1)
        # Attributes propagation
        if isinstance(self._base_type, XsdComplexType):
            for attr_use in self._base_type.attributes:
                attr_id = attr_use.referenced_node.uid
                if attr_id not in self.attr_ids:
                    self.attr_ids.add(attr_id)
                    self.attributes.append(attr_use)
                elif self._deriv_method == EXTENSION:
                    raise XsdError("%s already declared in %s" % (attr_id, self))
        self.derivation_resolved = True



# -*- coding: utf-8 -*-

from .errors import DtdError
from .utils import DtdNode, DtdUsage
from .constants import GROUP, ELEMENT


class DtdGroup(DtdNode):
    """
    Abstract class representing a group defined in a DTD.

    Two different groups can be defined: *choices* and *sequences*.
    A choice group defines several options from where one item must be chosen,
    a sequence group defines a sequence of items where the order matters.

    For this class, the ``category`` class attribute is set to ``GROUP``.
    """
    category = GROUP

    def __init__(self, name, uid_prefix=u""):
        super(DtdGroup, self).__init__(name, uid_prefix=uid_prefix)
        if self.__class__ is DtdGroup:
            raise NotImplementedError(u"Can't instantiate an abstract class")

    def add_item(self, item, min_occur=1, max_occur=1):
        raise NotImplementedError(u"Can't call abstract operations")

    def __len__(self):
        raise NotImplementedError(u"Can't call abstract operations")

    def __iter__(self):
        raise NotImplementedError(u"Can't call abstract operations")


class DtdChoice(DtdGroup):
    """
    Class representing a *choice* group defined in a DTD.

    A choice group defines several options from where one item must be chosen.
    """
    def __init__(self, name, uid_prefix=u""):
        super(DtdChoice, self).__init__(name, uid_prefix=uid_prefix)
        self.options = []
        self.option_ids = set()

    def add_item(self, item, min_occur=1, max_occur=1):
        """
        Adds the ``item`` item as one option into the *choice* group.

        This method checks that ``item`` category is ``ELEMENT`` or ``GROUP``,
        and that no other item with the same identifier exists in the group.
        """
        if item.category not in (ELEMENT, GROUP):
            raise DtdError("Can't add %s to %s (should be an element or a "
                           "group)" % (item, self))
        if item.uid in self.option_ids:
            raise DtdError("%s already declared in %s" % (item, self))
        self.option_ids.add(item.uid)
        self.options.append(DtdUsage(item, min_occur, max_occur))

    def __len__(self):
        return len(self.options)

    def __iter__(self):
        return iter(self.options)


class DtdSequence(DtdGroup):
    """
    Class representing a *choice* group defined in a DTD.

    A sequence group defines a sequence of items where the order
    matters.
    """
    def __init__(self, name, uid_prefix=u""):
        super(DtdSequence, self).__init__(name, uid_prefix=uid_prefix)
        self.children = []

    def add_item(self, item, min_occur=1, max_occur=1):
        """
        Adds the ``item`` item as one sub-item of the *sequence* group. The
        item is added at the end of the sequence.

        This method checks that ``item`` category is ``ELEMENT`` or ``GROUP``.
        """
        if item.category not in (ELEMENT, GROUP):
            raise DtdError("Can't add %s to %s (should be an element or a "
                           "group)" % (item, self))
        self.children.append(DtdUsage(item, min_occur, max_occur))

    def __len__(self):
        return len(self.children)

    def __iter__(self):
        return iter(self.children)



# -*- coding: utf-8 -*-

import os, sys
from os import path as osp
from lxml import etree
from .utils import DTD, XML_SCHEMA, split_qname


class XMLDocsGraphAnalyzer(object):
    def __init__(self, graph):
        self.title = graph.name
        self.graph = graph
        self.validation_method = u""
        self.validator=None
        if graph.source != u"" and graph.source is not None:
            self.validation_method = \
                        u"{0} {1}".format(graph.source_type, graph.source)
            filename = osp.abspath(graph.source)
            if graph.source_type == DTD:
                self.validator = etree.DTD(filename)
            elif graph.source_type == XML_SCHEMA:
                schema_tree = etree.parse(filename)
                self.validator = etree.XMLSchema(schema_tree)
            else:
                self.validation_method = u""
        self.val_results = {} # {filename: validation_result}
        self.graph.reset_uses()

    def analyze_xml_library(self, base_dir, suffixes=(u".xml",)):
        for path, sub_dirs, files in os.walk(base_dir):
            for filename in  [fn for fn in files
                              if osp.splitext(fn)[1] in suffixes]:
                filepath = osp.join(path, filename)
                self.analyze_xml_doc(base_dir, osp.relpath(filepath, base_dir))

    def analyze_xml_doc(self, base_dir, filename):
        if filename in self.val_results:
            return
        sys.stdout.write(u"analyzing {0}\n".format(filename))
        xml_tree = etree.parse(osp.join(base_dir, filename))
        if self.validator is not None:
            self.val_results[filename] = self.validator.validate(xml_tree)
        else:
            self.val_results[filename] = None
        if self.val_results[filename] is False:
            sys.stderr.write(
                u"### ERROR: {0} XML document is not valid against {1}\n===>"
                u" Skipping it\n".format(filename, self.validation_method))
            return
        self.visit_elt(xml_tree.getroot(), filename)

    def visit_elt(self, elt, filename, curr_gr_parent=None):
        elt_qname = self._get_elt_name(elt.tag)
        if curr_gr_parent is None:
            gr_elts = self.graph.find_high_level_elts(elt_qname)
        else:
            gr_elts = curr_gr_parent.find_children(elt_qname)
        if len(gr_elts) == 0:
            sys.stderr.write(
                u"### ERROR while analyzing {0}\nCan't find the graph "
                u"element corresponding to <{1}> inside the graph structure "
                u"describing the XML documents\n".format(filename, elt.tag))
            return
        elif len(gr_elts) > 1:
            sys.stderr.write(
                u"### WARNING while analyzing {0}\nFound more than one "
                u"graph element corresponding to <{1}> inside the graph "
                u"structure describing the XML documents\n===> Choosing the "
                u"first one\n".format(filename, elt.tag))
        gr_elt = gr_elts[0]
        gr_elt.actual_uses.setdefault(filename, 0)
        gr_elt.actual_uses[filename] += 1
        if curr_gr_parent is None:
            self.graph.root_uses[gr_elt].setdefault(filename, 0)
            self.graph.root_uses[gr_elt][filename] += 1
        else:
            curr_gr_parent.sub_elt_uses[gr_elt].setdefault(filename, 0)
            curr_gr_parent.sub_elt_uses[gr_elt][filename] += 1
        for attr_name in elt.attrib:
            attr_qname = self._get_attr_name(attr_name)
            gr_attrs = gr_elt.find_attributes(attr_qname)
            if len(gr_attrs) == 0:
                sys.stderr.write(
                    u"### ERROR while analyzing {0}\nCan't find the graph "
                    u"attribute corresponding to @{1} inside the graph "
                    u"structure describing the XML documents"
                    u"\n".format(filename, attr_name))
                continue
            elif len(gr_attrs) > 1:
                sys.stderr.write(
                    u"### ERROR while analyzing {0}\nFound more than one "
                    u"graph attribute corresponding to @{1} inside the "
                    u"graph structure describing the XML documents\n===> "
                    u"Choosing the first one\n".format(filename, attr_name))
            gr_attr = gr_attrs[0]
            gr_elt.attr_uses[gr_attr].setdefault(filename, 0)
            gr_elt.attr_uses[gr_attr][filename] += 1
        for child_elt in elt:
            if child_elt.tag is etree.Comment or child_elt.tag is etree.PI:
                continue
            self.visit_elt(child_elt, filename, curr_gr_parent=gr_elt)

    def _get_elt_name(self, elt_name):
        if self.graph.source_type == DTD and self.graph.dtd_qualif_local_elt:
            namespace, local_name = split_qname(elt_name)
            if namespace is None or namespace == u"":
                namespace = self.graph.dtd_ns_mapping.get(u"")
                if namespace is None:
                    return local_name
                else:
                    return u"{{{0}}}{1}".format(namespace, local_name)
        return elt_name

    def _get_attr_name(self, attr_name):
        if self.graph.source_type == DTD and self.graph.dtd_qualif_local_attr:
            namespace, local_name = split_qname(attr_name)
            if namespace is None or namespace == u"":
                namespace = self.graph.dtd_ns_mapping.get(u"")
                if namespace is None:
                    return local_name
                else:
                    return u"{{{0}}}{1}".format(namespace, local_name)
        return attr_name



# -*- coding: utf-8 -*-
"""
This module defines the class where the data read from a DTD will be stored.

.. autoclass:: DtdDefinition
   :show-inheritance: 
"""
__docformat__ = "restructuredtext en"


from .constants import ELEMENT
from .errors import DtdError

class DtdDefinition(object):
    """
    Class representing a DTD.

    This class stores the elements with their attributes and their sub-content.
    """
    def __init__(self):
        self.elements = {}

    def add_element(self, elt):
        """
        Adds an element described by a declaration from the DTD.

        This method checks that ``elt`` category is ``ELEMENT`` and that
        there is not already an element with the same identifier in this DTD.

        :param elt:
            Element to be stored ib the DTD.
        :type elt_decl:
            :class:`~dtd.nodes.Element`
        """
        if not elt.category is ELEMENT:
            raise DtdError(u"Can't add {0} to {1} as an element"
                           u"".format(elt, self))
        if elt.uid in self.elements:
            raise DtdError(u"{0} already declared in the DTD".format(elt.name))
        self.elements[elt.uid] = elt




#!/usr/bin/env python
# pylint: disable-msg=W0404,W0622,W0704,W0613,W0152
# copyright 2004-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <http://www.gnu.org/licenses/>.


from setuptools import setup, find_packages

long_description = open('README').read()


setup(name='pyxst',
      version='0.3.0',
      license='LGPL',
      description="XML Schema Tools for Python",
      long_description=long_description,
      long_description_content_type='text/x-rst',
      author='Logilab',
      author_email='devel@logilab.fr',
      url="https://www.logilab.org/project/pyxst",
      install_requires=[
          'lxml',
          'six'
      ],
      packages=find_packages(),
      )



# -*- coding: utf-8 -*-
"""
This module defines the classes that represent the groups read in an XML
Schema. The groups can be sequences, alls or choices.

The groups are defined as a specific node derived from
:class:`~pyxst.xml_struct.xsd.utils.XsdNode`.

.. autoclass:: XsdGroup
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: XsdChoice
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: XsdAll
   :members:
   :undoc-members:
   :show-inheritance: 

.. autoclass:: XsdSequence
   :members:
   :undoc-members:
   :show-inheritance: 
"""
__docformat__ = "restructuredtext en"


from .errors import XsdError
from .utils import XsdNode, XsdUsage
from .utils import GROUP, ELEMENT


class XsdGroup(XsdNode):
    """
    Abstract class representing a group defined in an XML Schema.

    Three different groups can be defined: *choices*, *alls* and *sequences*.
    A choice group defines several options from where one item must be chosen,
    an all group defines a set of items that can occur at most one time but in
    any order and a sequence group defines a sequence of items where the order
    matters.

    If a group is directly defined inside another group or inside a complex
    type, a name is nevertheless automatically created for this group.

    For this class, the ``category`` class attribute is set to ``GROUP``.

    .. automethod:: __init__
    """
    category = GROUP

    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new group.

        :param name:
            Name of the group.
        :type name:
            :class:`unicode`
        """
        super(XsdGroup, self).__init__(name, uid_prefix=uid_prefix)
        if self.__class__ is XsdGroup:
            raise NotImplementedError("Can't instantiate an abstract class")


class XsdChoice(XsdGroup):
    """
    Class representing a *choice* group defined in an XML Schema.

    A choice group defines several options from where one item must be chosen.

    .. attribute:: options 

       List of all the options defined in this *choice* group. Only one of 
       these options must be chosen in the XML document.

       The uniqueness of each option is checked thanks to the ``option_ids``
       attribute of this class.

       Type: list of 
       :class:`~pyxst.xml_struct.xsd.element.XsdElement` or
       :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`

    .. attribute:: option_ids

       Set containing the identifiers of the options of this group. Hence, 
       this group can't contain two items with the same name.

       This attribute is only used to check that new options added
       to this group don't collide with already defined options with the same 
       name.

       Type: set of :class:`str`

    .. automethod:: __init__
    .. automethod:: __len__
    .. automethod:: __iter__
    """
    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new *choice* group.

        A *choice* group is a group where only one sub-item can occur chosen
        among the items defined in the *choice* group.

        :param name:
            Name of the group.
        :type name:
            :class:`unicode`
        """
        super(XsdChoice, self).__init__(name, uid_prefix=uid_prefix)
        self.options = []
        self.option_ids = set()

    def add_item(self, item, min_occur=1, max_occur=1):
        """
        Adds the ``item`` item as one option into the *choice* group.

        This method checks that ``item`` category is ``ELEMENT`` or ``GROUP``,
        and that no other item with the same identifier exists in the group.

        :param item:
            Item node to be added in this group.
        :type attr:
            :class:`~pyxst.xml_struct.xsd.elements.XsdElement` or
            :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        :param min_occur:
            Minimum occurence for the item
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence for the item
        :type max_occur:
            :class:`int` or 
            :data:`~pyxst.xml_struct.utils.INFINITY`
        """
        item_id = item.uid
        if item.category not in [ELEMENT, GROUP]:
            raise XsdError("Can't add %s to %s (should be an element or a "
                           "group)" % (item, self))
        if item_id in self.option_ids:
            raise XsdError("%s already declared in %s" % (item, self))
        self.option_ids.add(item_id)
        self.options.append(XsdUsage(item, min_occur, max_occur))

    def __len__(self):
        """
        Returns the number of options defined in this group.

        :return:
            The number of different options that can occur in the group.
        :rtype:
            :class:`int`
        """
        return len(self.options)

    def __iter__(self):
        """
        Returns an iterator over the different options that can occur in this 
        group.

        :return:
            An iterator over the various options defined in the group.
        :rtype:
            :class:`listiterator` over 
            :class:`~pyxst.xml_struct.xsd.utils.XsdUsage`
        """
        return iter(self.options)


class XsdAll(XsdGroup):
    """
    Class representing an *all* group defined in an XML Schema.

    An all group defines a set of items that can occur at most one time but in
    any order.

    .. attribute:: children 

       List of all the children defined in this *all* group. At most
       one occurence of each child can be inserted in the XML
       document, but in any order.

       The uniqueness of each child is checked thanks to the ``child_ids``
       attribute of this class.

       Type: list of 
       :class:`~pyxst.xml_struct.xsd.elements.XsdElement` or
       :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`

    .. attribute:: child_ids

       Set containing the identifiers of the children of this group. Hence, 
       this group can't contain two items with the same name.

       This attribute is only used to check that new children added
       to this group don't collide with already defined children with the same 
       name.

       Type: set of :class:`str`

    .. automethod:: __init__
    .. automethod:: __len__
    .. automethod:: __iter__
    """
    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new *all* group.

        An *all* group is a group where the sub-items can occur in any order 
        but at most one time.

        :param name:
            Name of the group.
        :type name:
            :class:`unicode`
        """
        super(XsdAll, self).__init__(name, uid_prefix=uid_prefix)
        self.children = []
        self.child_ids = set()

    def add_item(self, item, min_occur=1, max_occur=1):
        """
        Adds the ``item`` item as one sub-item of the *all* group.

        This method checks that ``item`` category is ``ELEMENT`` or ``GROUP``,
        that no other item with the same identifier exists in the group, and
        that the maximum occurence is not greater than 1.

        :param item:
            Item node to be added in this group.
        :type attr:
            :class:`~pyxst.xml_struct.xsd.elements.XsdElement` or
            :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        :param min_occur:
            Minimum occurence for the item (0 or 1)
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence for the item (0 or 1)
        :type max_occur:
            :class:`int`
        """
        item_id = item.uid
        if item.category not in [ELEMENT, GROUP]:
            raise XsdError("Can't add %s to %s (should be an element or a "
                           "group)" % (item, self))
        if item_id in self.child_ids:
            raise XsdError("%s already declared in %s" % (item, self))
        if max_occur > 1:
            raise XsdError("%s can't occur more than one time in %s"
                           % (item, self))
        self.child_ids.add(item_id)
        self.children.append(XsdUsage(item, min_occur, max_occur))

    def __len__(self):
        """
        Returns the number of children defined in this all.

        :return:
            The number of different children that can occur in the all.
        :rtype:
            :class:`int`
        """
        return len(self.children)

    def __iter__(self):
        """
        Returns an iterator over the children that can occur in this 
        group.

        :return:
            An iterator over the children defined in the group.
        :rtype:
            :class:`listiterator` over 
            :class:`~pyxst.xml_struct.xsd.utils.XsdUsage`
        """
        return iter(self.children)


class XsdSequence(XsdGroup):
    """
    Class representing a *choice* group defined in an XML Schema.

    A sequence group defines a sequence of items where the order
    matters.

    .. attribute:: children 

       List of all the children defined in this *sequence* group. Each child
       must be inserted in the XML document in the exact order defined in this
       group.

       Type: list of 
       :class:`~pyxst.xml_struct.xsd.elements.XsdElement` or
       :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`

    .. automethod:: __init__
    .. automethod:: __len__
    .. automethod:: __iter__
    """
    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new *sequence* group.

        A *sequence* group is a group where the sub-items must respect
        the given order but can occur any number of times.

        :param name:
            Name of the group.
        :type name:
            :class:`unicode`
        """
        super(XsdSequence, self).__init__(name, uid_prefix=uid_prefix)
        self.children = []

    def add_item(self, item, min_occur=1, max_occur=1):
        """
        Adds the ``item`` item as one sub-item of the *sequence* group. The
        item is added at the end of the sequence.

        This method checks that ``item`` category is ``ELEMENT`` or ``GROUP``.

        :param item:
            Item node to be added in this group.
        :type attr:
            :class:`~pyxst.xml_struct.xsd.elements.XsdElement` or
            :class:`~pyxst.xml_struct.xsd.groups.XsdGroup`
        :param min_occur:
            Minimum occurence for the item
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence for the item
        :type max_occur:
            :class:`int` or 
            :data:`~pyxst.xml_struct.utils.INFINITY`
        """
        if item.category not in [ELEMENT, GROUP]:
            raise XsdError("Can't add %s to %s (should be an element or a "
                           "group)" % (item, self))
        self.children.append(XsdUsage(item, min_occur, max_occur))

    def __len__(self):
        """
        Returns the number of children defined in this sequence.

        :return:
            The number of different children that can occur in the sequence.
        :rtype:
            :class:`int`
        """
        return len(self.children)

    def __iter__(self):
        """
        Returns an iterator over the children that can occur in this 
        group.

        :return:
            An iterator over the children defined in the group.
        :rtype:
            :class:`listiterator` over 
            :class:`~pyxst.xml_struct.xsd.utils.XsdUsage`
        """
        return iter(self.children)



# -*- coding: utf-8 -*-
"""
This module defines several utilitary constants and functions, mainly for
proessing the names and namespaces of XML nodes.

Constants
.........

.. autodata:: INFINITY

.. autodata:: ANY

.. autodata:: MIXED

.. autodata:: UNKNOWN

.. autodata:: BOOLEAN

.. autodata:: INTEGER

.. autodata:: FLOAT

.. autodata:: STRING

.. autodata:: DATETIME

.. autodata:: DURATION

.. autodata:: BASE64

.. autodata:: HEXA

.. autodata:: ABSENT
.. autodata:: DIFFERENT
.. autodata:: SAME

.. autodata EMPTY
.. autodata TEXT_ONLY
.. autodata ELEMENTS_ONLY

.. autodata DTD
.. autodata XML_SCHEMA

Functions
.........

.. autofunction:: split_qname

.. autofunction:: combine_types
"""
__docformat__ = "restructuredtext en"


INFINITY = u"Infinity"
"""
Constant describing the infinity (for unbounded occurences).

Type: :class:`unicode`
"""

ANY = u"Any text or elements"
"""
Constant describing the type of content that can be anything.

Type: :class:`unicode`
"""

MIXED = u"Mixed Content"
"""
Constant describing the type of the content that can occur inside an element
if this element can contain, at the same time, sub-elements and textual data.

Type: :class:`unicode`
"""

UNKNOWN = u"Unknown"
"""
Constant describing the type of a textual content that is not known.

Type: :class:`unicode`
"""

BOOLEAN = u"Boolean"
"""
Constant describing the type of a textual content that is a boolean (``true``
or ``false`` textual value).

Type: :class:`unicode`
"""

INTEGER = u"Integer"
"""
Constant describing the type of a textual content that is an integer.

Type: :class:`unicode`
"""

FLOAT = u"Float"
"""
Constant describing the type of a textual content that is a real (float).
The ``.`` character is the separator between integer and decimal parts.

Type: :class:`unicode`
"""

STRING = u"String"
"""
Constant describing the type of a textual content that is a string (text).

Type: :class:`unicode`
"""

DATETIME = u"DateTime"
"""
Constant describing the type of a textual content that is a date, a
time or a date-time. The text representation of a date-time is defined in the
XML Schema standard.

Type: :class:`unicode`
"""

DURATION = u"Duration"
"""
Constant describing the type of a textual content that is a duration between
two dates-times. The text representation of a duration is defined in the
XML Schema standard.

Type: :class:`unicode`
"""

BASE64 = u"Base64"
"""
Constant describing the type of a textual content that is a binary content
coded in base 64.

Type: :class:`unicode`
"""

HEXA = u"Hexadecimal"
"""
Constant describing the type of a textual content that is a binary content
coded in hexadecimal.

Type: :class:`unicode`
"""

EMPTY = u"Empty"
TEXT_ONLY = u"Text only"
ELEMENTS_ONLY = u"Elements only"

ABSENT = u"Absent item"
DIFFERENT = u"Different item"
SAME = u"Same item"

DTD = u"DTD"
XML_SCHEMA = u"XML Schema"


def split_qname(qname):
    """
    Splits the XML qualified name in a tuple containing the namespace and the
    local name.

    A qualified name is composed with the namespace and the local name.
    In the XML nodes read with :mod:`xml.etree` module, the qualified name
    is built as follow: ``{namespace}local-name``.

    :param qname:
        XML qualified name as built by the :mod:`xml.etree` module.
    :type qname:
        :class:`unicode`
    :returns:
        The namespace and the local name extracted from ``qname``
    :rtype:
        tuple composed with :class:`unicode` and :class:`unicode`
    """
    if qname is None:
        return None, None
    name = str(qname)
    if name.find('}') == -1:
        return None, name
    return name[name.find('{') + 1:name.find('}')], name[name.find('}') + 1:]



# -*- coding: utf-8 -*-
"""
This module defines the classes that represent the elements read in an XML
Schema.

The elements are defined as a specific node derived from
:class:`~pyxst.xml_struct.xsd.utils.XsdNode`.

.. autoclass:: XsdElement
   :members:
   :undoc-members:
   :show-inheritance: 
"""
__docformat__ = "restructuredtext en"


from .utils import XsdNode, ELEMENT


class XsdElement(XsdNode):
    """
    Class representing an element defined in an XML Schema.

    .. attribute:: xsd_type 

       Type of this element declared in the XML Schema. This type is a
       type node. The type of an element can be left undefined
       when creating the element and defined later. When the XML Schema has
       been read, all the elements must have a type.

       Type: :class:`~pyxst.xml_struct.xsd.types.XsdType`

    .. attribute:: _elts_with_same_type

       Private attribute used to temporarly store other elements that
       have the same type as this element. When the type of this
       element is defined, it must also be given to all these other
       elements. At the end of the building step, this attribute is empty as
       each element directly points to its type.

       Type: List of 
       :class:`~pyxst.xml_struct.xsd.elements.XsdElement`

    For this class, the ``category`` class attribute is set to ``ELEMENT``.

    .. automethod:: __init__
    """
    category = ELEMENT

    def __init__(self, name, uid_prefix=u""):
        """
        Initializes a new element node.

        :param name:
            Name of the element.
        :type name:
            :class:`unicode`
        """
        super(XsdElement, self).__init__(name, uid_prefix=uid_prefix)
        self.xsd_type = None
        self._elts_with_same_type = []



# -*- coding: utf-8 -*-

ATTRIBUTE = u"Attribute"
"""
Constant containing the category of the attribute nodes

Type: :class:`unicode`
"""

ELEMENT = u"Element"
"""
Constant containing the category of the element nodes

Type: :class:`unicode`
"""

GROUP = u"Group"
"""
Constant containing the category of the group nodes

Type: :class:`unicode`
"""

INFINITY = u"infinity"
"""
Constant describing the infinity (for unbounded occurences).

Type: :class:`unicode`
"""

EMPTY = u"Empty"
"""
Constant describing the type of the content that can occur inside an element
if this element contain nothing.

Type: :class:`unicode`
"""

ELEMENTS_ONLY = u"Elements only"
"""
Constant describing the type of the content that can occur inside an element
if this element contain only sub-elements.

Type: :class:`unicode`
"""

TEXT_ONLY = u"Text only"
"""
Constant describing the type of the content that can occur inside an element
if this element contain only text data.

Type: :class:`unicode`
"""

MIXED = u"Mixed Content"
"""
Constant describing the type of the content that can occur inside an element
if this element can contain, at the same time, sub-elements and textual data.

Type: :class:`unicode`
"""

ANY = u"Any text or element"
"""
Constant describing the type of the content that can occur inside an element
if this element can contain anything (any text data and any elements).

Type: :class:`unicode`
"""



# -*- coding: utf-8 -*-
"""This package contains classes and functions that
produce a graph representing the XML structure described in the schema.

The :mod:`~xml_struct.graph` defines the items used to build the
actual structure that will be returned by this package. This structure
is a graph describing the structure of the XML documents (XML elements
that contain XML attributes and XML sub-elements).

The function defined in this module can read a
schema and return a grah describing the XML structure.

.. autofunction:: build_xml_structure_from_schema

"""
__docformat__ = "restructuredtext en"

from os.path import split, splitext


from ..xsd import read_schema
from ..dtd import read_dtd
from .utils import DTD, XML_SCHEMA
from .graph import XMLGraph
from .graph_target_relations import weave_target_relations


def build_xml_structure_from_schema(xsd_file, type_mapper=None):
    """
    Reads the ``xsd_file``, parses the XML Schema and builds a graph describing
    the structure of the XML data.

    The graph is an :class:`~xml_struct.graph.XMLGraph` that contains
    nodes describing the various elements, their attributes and their
    inner content (textual data or structure of sub-elements).

    :param xsd_file:
        Name of the file that contains the XML Schema.
    :type xsd_file:
        :class:`str`
    :returns:
        A graph describing the structure of the XML data.
    :rtype:
        :class:`~xml_struct.graph.XMLGraph`
    """
    try:
        name = splitext(split(xsd_file)[1])[0]
    except Exception:
        name = "anonymous"
    schema = read_schema(xsd_file)
    graph = XMLGraph(name, xsd_file, XML_SCHEMA, type_mapper)
    graph.build_from_schema(schema)
    weave_target_relations(graph)
    return graph


def build_xml_structure_from_dtd(dtd_file, ns_mapping=None,
                                 qualify_local_elements=False,
                                 qualify_local_attributes=False):
    """
    Reads the ``dtd_file``, parses the DTD and builds a graph describing
    the structure of the XML data.

    The graph is an :class:`~xml_struct.graph.XMLGraph` that contains
    nodes describing the various elements, their attributes and their
    inner content (textual data or structure of sub-elements).

    The DTD doesn't support namespaces. The elements and attributes have to
    be declared with their prefix (e.g. ``xlink:href``). The ``ns_mapping``
    parameter allows the building of qualified names (namespace plus local
    name) from the DTD declaration.

    :param dtd_file:
        Name of the file that contains the DTD.
    :type dtd_file:
        :class:`unicode`
    :param ns_mapping:
        Mapping of prefixes towards namespaces ``{prefix: namespace}``. It is
        possible to give a default namespace by associating a namespace with
        ``u""`` or ``None``.
    :type ns_mapping:
        dict {:class:`unicode`: :class:`unicode`}
    :param qualify_local_elements:
        Flag indicating that the elements without a prefix must be qualified
        with the default namespace given in ``ns_mapping`` (namespace
        indexed with ``u""``).
    :type qualify_local_elements:
        :class:`bool`
    :param qualify_local_attributes:
        Flag indicating that the attributes without a prefix must be qualified
        with the default namespace given in ``ns_mapping`` (namespace
        indexed with ``u""``).
    :type qualify_local_attributes:
        :class:`bool`
    :returns:
        A graph describing the structure of the XML data.
    :rtype:
        :class:`~pyxst.xml_struct.graph.XMLGraph`
    """
    try:
        name = splitext(split(dtd_file)[1])[0]
    except Exception:
        name = "anonymous"
    dtd_def = read_dtd(dtd_file, ns_mapping, qualify_local_elements,
                       qualify_local_attributes)
    graph = XMLGraph(name, dtd_file, DTD)
    graph.dtd_ns_mapping = ns_mapping or {}
    graph.dtd_qualif_local_elt = qualify_local_elements
    graph.dtd_qualif_local_attr = qualify_local_attributes
    graph.build_from_dtd(dtd_def)
    weave_target_relations(graph)
    return graph



# -*- coding: utf-8 -*-
"""
This module defines the classes used to build the graph of the actual
structure that describes the XML documents.

These classes are the nodes that compose the graph defined with :class:`~pyxst.xml_struct.graph.XMLGraph` that will be returned by :mod:`~pyxst.xml_struct` module.

These nodes are mainly XML elements and XML attributes that are
connected with links that give the minimum and maximum numbers of
occurence of the child nodes inside their parent node.

The structure created with these nodes is simpler than the
structure read in the XML Schemas with the
:mod:`pyxst.xml_struct.xsd` module.

Utilitary classes
.................

.. autoclass:: XMLNode
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: Occurence
   :members:
   :undoc-members:
   :show-inheritance:

Groups of child nodes
.....................

.. autoclass:: Group
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: Sequence
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: Alternative
   :members:
   :undoc-members:
   :show-inheritance:

Actual XML nodes
................

.. autoclass:: XMLElement
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: XMLAttribute
   :members:
   :undoc-members:
   :show-inheritance:
"""
__docformat__ = "restructuredtext en"


import operator
from .utils import (split_qname, INFINITY, EMPTY, ELEMENTS_ONLY,
                              TEXT_ONLY, MIXED, UNKNOWN, ANY)


class XMLNode(object):
    """
    Abstract class that represents an XML node, ie either an element or an
    attribute.

    This class defines common attributes.

    .. attribute:: local_name

       Local name of the XML node (name without namespace). The
       namespace and the local name compose the qualified name.

       Type: :class:`unicode`

    .. attribute:: namespace

       Namespace of the XML node. The namespace and the local name compose the
       qualified name.

       Type: :class:`unicode`

    .. attribute:: parent_elts

       List of XML elements that can be parents of this XML node.

       The attributes have a single parent. The local elements also have
       a single parent. The high level elements can occur in several elements
       and therefore can have several possible parents.

       The order in the list of parents is not relevant.

       Type: list of :class:`~pyxst.xml_struct.graph_nodes.XMLElement`

    .. attribute:: textual_content_type

       Type of the textual content that can be found in this element. The
       types are basic types such as integer, float, date, datetime, duration,
       string.

       If the XML node is an element that have mixed content ie a content that
       mixes child elements and textual data, this type is set to ``MIXED``.

       If the XML node is an empty element or has only elements, this type is
       set to ``None``.

       Type: :class:`str`

    .. attribute:: textual_content_is_list

       Flag set to true if the textual content is a space-separated list of
       items whose type is given by ``textual_content_type`` attribute.

       Type: :class:`bool`

    .. attribute:: textual_content_values

       Set containing the possible values for the textual content
       (when this content is contrained). If this attribute contains ``None``,
       the value of the textual content is free.

       Type: :class:`set` of :class:`unicode`

    .. automethod:: __init__
    """
    parent = None

    def __init__(self, qname):
        """
        Initializes a new XML node.

        :param qname:
            Qualified name of the XML node (namespace + local name). The
            qualified name is expected to be in the ``{namespace}local-name``
            form.
        :type qname:
            :class:`unicode`
        """
        namespace, local_name = split_qname(qname)
        self.local_name = local_name
        self.namespace = namespace
        self.parent_elts = []
        self.textual_content_type = None
        self.textual_content_is_list = False
        self.textual_content_values = None
        if self.__class__ is XMLNode:
            raise NotImplementedError("Can't instantiate abstract classes")

    @property
    def qname(self):
        """
        Gives the qualified name of the XML node.

        :returns:
            The qualified name of the XML node in the
            ``{namespace}local-name`` form.
        :rtype:
            :class:`unicode`
        """
        if self.namespace is None:
            return self.local_name
        else:
            return "{%s}%s" % (self.namespace, self.local_name)

    def add_parent_elt(self, parent):
        """
        Adds a parent element for this node.

        **Abstract method that will be defined in derived classes.**

        :param parent:
            Parent element of this node (element that contains the node).
        :type parent:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        """
        raise NotImplementedError("Can't call abstract mathod")

    def __repr__(self):
        return u"Node <{0}>".format(self.qname)

class Occurence(object):
    """
    Class that describes a link between a parent node and its child node.

    This class gives the minimum and maximum occurence of the child node
    inside its parent node. It will appear in the attributes list of
    an element or in the children list of a group.

    .. attribute:: target

       Child node.

       Type: :class:`~pyxst.xml_struct.graph_nodes.XMLNode`

    .. attribute:: minimum

       Minimum occurence of the child node inside its parent node.

       Type: :class:`int`

    .. attribute:: maximum

       Maximum occurence of the child node inside its parent node.

       Can be ``INFINITY`` if there is no maximum.

       Type: :class:`int` or
       :data:`~pyxst.xml_struct.utils.INFINITY`

    .. automethod:: __init__
    """

    def __init__(self, target, min_occur, max_occur):
        """
        Initializes a new occurence.

        :param target:
            Child node that this occurence points to.
        :type target:
            :class:`~pyxst.xml_struct.graph_nodes.XMLNode`
        :param min_occur:
            Minimum occurence of the child node inside its parent.
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence of the child node inside its parent. Can be
            ``INFINITY``
        :type max_occur:
            :class:`int` or
            :data:`~pyxst.xml_struct.utils.INFINITY`
        """
        self.target = target
        target.parent = self
        self.minimum = min_occur
        self.maximum = max_occur

    @property
    def qname(self):
        return self.target.qname

    def add_parent_elt(self, parent):
        """
        Adds a parent element on the child referenced by this occurence object.

        :param parent:
            Parent element to be added to the child pointed by this occurence
        :type parent:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        """
        self.target.add_parent_elt(parent)

    def simplify(self):
        return self.target.simplify(self.minimum, self.maximum)

    def __eq__(self, other):
        if not isinstance(other, Occurence):
            return NotImplemented
        if isinstance(self.target, XMLElement):
            if not isinstance(other.target, XMLElement):
                return False
            return (self.target.qname == other.target.qname
                    and self.minimum == other.minimum
                    and self.maximum == other.maximum)
        if isinstance(self.target, Group):
            if self.target.__class__ != other.target.__class__:
                return False
            return (self.minimum == other.minimum
                    and self.maximum == other.maximum)
        return (self.target == other.target
                and self.minimum == other.minimum
                and self.maximum == other.maximum)

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        if self.minimum == self.maximum:
            occurs = u"[{0}]".format(self.minimum)
        else:
            occurs = u"[{0},{1}]".format(self.minimum, self.maximum)
        return u"{0} {1}".format(self.target, occurs)


class Group(object):
    """
    Class describing a group of XML elements that are children of a parent
    element. This group is a non-ordered group (the child elements can occur
    in any order).

    A group can directly contain child elements or can contain other groups
    that contain child elements.

    .. attribute:: children

       Child elements or groups that occur inside this group. Whereas this
       attribute is a list, the order of the child elements is not constrained.
       This attribute contains ``Occurence`` instances that links towards the
       actual child (that can be an element or another group) and keeps its
       minimum occurence and its maximum occurence.

       Type: list of
       :class:`~pyxst.xml_struct.graph_nodes.Occurence`

    .. attribute:: parent_elt

       Parent element that contains this group.

       If this group is a sub-group that is contained in another group, the
       parent element is the parent element of the highest level group.

       A group is always dedicated to a single element and therefore has
       a single parent element.

       Type: :class:`~pyxst.xml_struct.graph_nodes.XMLElement`

    .. automethod:: __init__

    .. automethod:: __len__

    .. automethod:: __getitem__
    """
    def __init__(self, parent=None):
        """
        Initializes a new group.If this group is
            contained in another group, the parent element is the parent
            element of the highest level group.

        :param parent:
            Parent element that contains this group.
        :type parent:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        """
        self.children = []
        self.parent_elt = parent

    def __len__(self):
        """
        Returns the length of the group ie the number of children.

        :returns:
            The number of children.
        :rtype:
            :class:`int`
        """
        return len(self.children)

    def __getitem__(self, idx):
        """
        Returns a tuple containing the occurence of the child in position
        number ``idx``.

        Even if the group is not ordered, it is possible to iterate over the
        children.

        :param idx:
            Index of the child to be returned.
        :param idx:
            :class:`int`
        :returns:
            The child in idx-th position and its minimum and maximum occurences
        :rtype: :class:`~pyxst.xml_struct.graph_nodes.Occurence
        """
        return self.children[idx]

    def add_child(self, child, min_occur=1, max_occur=1):
        """
        Adds a new child in this group.

        If the group is ordered, the child is appended after the existing
        children. This method automatically creates an ``Occurence`` instance
        to keep the minimum and maximum occurence inside the link towards the
        child.

        :param child:
            Child to be appended in the group.
        :type child:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement` or
            :class:`~pyxst.xml_struct.graph_nodes.Group`
        :param min_occur:
            Minimum occurence of the child inside this group.
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence of the child inside this group.
        :type max_occur:
            :class:`int`
        """
        self.children.append(Occurence(child, min_occur, max_occur))
        if self.parent_elt is not None:
            child.add_parent_elt(self.parent_elt)

    def add_parent_elt(self, parent):
        """
        Adds a parent element for this group and for all its children.

        :param parent:
            Parent element to be added to this group and all its children.
        :type parent:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        """
        if self.parent_elt is not None and self.parent_elt != parent:
            raise Exception("Groups can't have two different parents")
        self.parent_elt = parent
        for child in self.children:
            child.add_parent_elt(parent)

    def simplify(self, min_occur=None, max_occur=None):
        """
        Simplifies the group and its sub-groups (if any).

        Before working on this group, this method recursively calls
        the ``simplify`` method of the sub-groups.

        If a child (element or sub-group) has a maximum occurence
        equaled to ``0``, this method deletes it.

        If a sub-group is empty, this method deletes it.

        If a sub-group has only one child, directly adds the child
        inside this group.

        If a sub-group has the same sort as this group (same class)
        and the maximum occurence of this sub-group is ``1`` and the
        minimum occurence of the children of this sub-group is never
        more than``1``, this method merges this sub-group into this
        group.
        """
        change = False
        for occur in self.children[:]:
            child = occur.target
            child_min = occur.minimum
            child_max = occur.maximum
            if child_max == 0:
                self.children.remove(occur)
                change = True
            elif isinstance(child, Group):
                change |= occur.simplify()
                if len(child) == 0:
                    self.children.remove(occur)
                    change = True
                elif len(child) == 1:
                    occ = child.children.pop()
                    occ.minimum = child_min * occ.minimum
                    if child_max == INFINITY or occ.maximum == INFINITY:
                        occ.maximum = INFINITY
                    else:
                        occ.maximum = child_max * occ.maximum
                    idx = self.children.index(occur) + 1
                    self.children.insert(idx, occ)
                    self.children.remove(occur)
                    change = True
                elif child.__class__ is self.__class__ and child_max == 1:
                    can_simplify = True
                    for occ in child:
                        sub_min = occ.minimum
                        can_simplify &= (sub_min <= 1)
                    if can_simplify:
                        idx = self.children.index(occur) + 1
                        for i in range(len(child.children)):
                            occ = child.children.pop()
                            occ.minimum *= child_min
                            self.children.insert(idx, occ)
                        self.children.remove(occur)
                        change = True
        return change

    def sort(self):
        """
        Sorts the group children and recursively calls the sort method of
        the sub-groups.
        """
        sub_groups = []
        sub_elts = []
        for occur in self.children:
            if isinstance(occur.target, Group):
                occur.target.sort()
                sub_groups.append(occur)
            else:
                sub_elts.append(occur)
        sub_elts.sort(key=operator.attrgetter("qname"))
        self.children = sub_elts + sub_groups

    def list_sub_elts(self, recursive):
        """
        Returns a list with all the XML elements that appear in this group
        and its sub-groups (if recursive is True).
        """
        result = []
        for occur in self.children:
            if isinstance(occur.target, XMLElement):
                result.append(occur.target)
            elif recursive and isinstance(occur.target, Group):
                result.extend(occur.target.list_sub_elts(recursive))
        return result

    def __repr__(self):
        results = []
        for occ in self.children:
            results.append(u"{0}".format(occ))
        return u"Group({0})".format(u", ".join(results))


class Sequence(Group):
    """
    Class describing an ordered group of children. The children can be either
    elements or sub-groups.

    .. automethod:: __init__
    """
    def __init__(self, parent=None):
        """
        Initializes a new ordered group.
        """
        super(Sequence, self).__init__(parent)

    def sort(self):
        """
        Doesn't sort the group children but recursively calls the sort
        method on the sub-groups.
        """
        for occur in self.children:
            if isinstance(occur.target, Group):
                occur.target.sort()

    def __repr__(self):
        results = []
        for occ in self.children:
            results.append(u"{0}".format(occ))
        return u"Sequence({0})".format(u", ".join(results))

class Alternative(Group):
    """
    Class describing an alternative between several children. In the actual
    XML document, only one of the possible child will occur.

    The children can be either elements or sub-groups.

    .. automethod:: __init__
    """
    def __init__(self, parent=None):
        """
        Initializes a new alternative group.
        """
        super(Alternative, self).__init__(parent)

    def simplify(self, min_occur=None, max_occur=None):
        """
        Simplifies the alternative and its sub-groups (if any).

        First, calls the ``simplify`` method on the base class.

        If the maximum occurrence of this alternative (``max_occur``) is
        INFINITY, the maximum occurence of its children can be the same
        as their minimum occurence (1 if their minimum is 0).

        If the minimum occurrence of this alternative (``min_occur``) is
        0, the minimum occurence of its children can be 1 instead of 0.
        """
        change = super(Alternative, self).simplify(min_occur, max_occur)
        if max_occur == INFINITY:
            for occ in self.children:
                if occ.maximum == INFINITY or occ.maximum > occ.minimum:
                    if occ.minimum == 0:
                        occ.maximum = 1
                    else:
                        occ.maximum = occ.minimum
                    change = True
        if min_occur == 0:
            for occ in self.children:
                if occ.minimum == 0:
                    occ.minimum = 1
                    change = True
        return change

    def __repr__(self):
        results = []
        for occ in self.children:
            results.append(u"{0}".format(occ))
        return u"Alternative({0})".format(u" | ".join(results))

class XMLElement(XMLNode):
    """
    Class defining an XML element.

    .. attribute:: is_high_level

       Flag set to true if the element is a high level element.

       High level elements are elements that might be the root element
       of the XML document. When the XML structure is read from an XML
       Schema, all the elements defined at the top level of the schema
       are high level elements. When the XML structure is read from a
       DTD, all the elements are high level elements.

       Type: :class:`bool`

    .. attribute:: content

       Occurence towards the group that contains the child elements of this
       element.

       This content can be ``None`` if this element is empty.

       Type: :class:`~pyxst.xml_struct.graph_nodes.Occurence`

    .. attribute:: attributes

       Dictionary grouping the occurences towards the attributes of
       this element. The attributes are indexed with their qualified
       names.

       Type: dictionary of
       :class:`~pyxst.xml_struct.graph_nodes.Occurence` indexed
       with :class:`unicode`.

    .. automethod:: __init__
    """
    def __init__(self, qname, is_high_level=False, desc=None):
        """
        Initializes a new XML element.

        :param qname:
            Qualified name of the XML node (namespace + local
            name). The qualified name is expected to be in the
            ``{namespace}local-name`` form.
        :type qname:
            :class:`unicode`
        :param is_high_level:
            Flag indicating if the element is a high level element.
        :type is_high_level:
            :class:`bool`
        """
        super(XMLElement, self).__init__(qname)
        self.is_high_level = is_high_level
        self.desc = desc
        self.content_type = UNKNOWN
        self.content = None
        self.attributes = {}
        # The two following attributes are used for analyzing a library of XML
        # documents whose structure is given by the graph. They are used to
        # count the number of times an attribute or a sub-element is actually
        # used in all the XML documents of the library.
        # (Some elements or attributes might never be used in the documents)
        self.actual_uses = {} # {xml_filename: uses_count,}
        self.attr_uses = {} # {XMLAttribute: {xml_filename: uses_count,}}
        self.sub_elt_uses = {} # {XMLElement: {xml_filename: uses_count,}}

    def get(self, attr_qname):
        """
        Returns a tuple containing the attribute named ``attr_qname``,
        its minimum occurence and its maximum occurence.

        :param attr_qname:
            Qualified name of the attribute to be returned.
        :param idx:
            :class:`unicode`
        :returns:
            The attribute whose name is ``attr_qname`` and its minimum and
            maximum occurences
        :rtype:
            Tuple composed with
            :class:`~pyxst.xml_struct.graph_nodes.XMLAttribute`,
            :class:`int`, :class:`int`
        """
        occ = self.attributes[attr_qname]
        return occ.target, occ.minimum, occ.maximum

    def iter_attrs(self):
        return sorted(self.attributes.values(),
                      key=operator.attrgetter("qname"))

    def add_attribute(self, attr, min_occur=1, max_occur=1):
        """
        Adds a new attribute in this element.

        This method automatically creates an ``Occurence`` instance to
        keep the minimum and maximum occurence inside the link towards
        the attribute.

        :param attr:
            Attribute to be added in the element.
        :type attr:
            :class:`~pyxst.xml_struct.graph_nodes.XMLAttribute`
        :param min_occur:
            Minimum occurence of the attribute inside this element (0 or 1).
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence of the attribute inside this element (1).
        :type max_occur:
            :class:`int`
        """
        self.attributes[attr.qname] = \
            Occurence(attr, min_occur, max_occur)
        attr.add_parent_elt(self)

    def set_content(self, group, min_occur=1, max_occur=1):
        """
        Sets ``group`` as the content that describes the child elements of this
        element.

        :param group:
            Group that describes the child elements that can occur inside this
            element.
        :type group:
            :class:`~pyxst.xml_struct.graph_nodes.Group`
        :param min_occur:
            Minimum occurence of the attribute inside this element (0 or 1).
        :type min_occur:
            :class:`int`
        :param max_occur:
            Maximum occurence of the attribute inside this element (1).
        :type max_occur:
            :class:`int`
        """
        self.content = Occurence(group, min_occur, max_occur)
        group.add_parent_elt(self)

    def add_parent_elt(self, parent):
        """
        Adds a parent element for this element.

        :param parent:
            Parent element of this element (element that can contain this
            element).
        :type parent:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        """
        if parent not in self.parent_elts:
            self.parent_elts.append(parent)

    def compute_content_type(self):
        if self.textual_content_type is None and self.content is None:
            self.content_type = EMPTY
        elif self.textual_content_type is None:
            self.content_type = ELEMENTS_ONLY
        elif self.textual_content_type in (MIXED, UNKNOWN, ANY):
            self.content_type = self.textual_content_type
        else:
            self.content_type = TEXT_ONLY

    def simplify(self):
        """
        Calls the ``simplify`` method on the group that describes the content
        of the element (if any).

        The ``simplify`` method erases empty sub_groups, merges
        sub-groups with their parent group when possible and corrects
        the occurences inside the alternative groups when applicable.

        If the top-level group in this element is empty, this
        method deletes it. It also deletes the attributes whose
        maximum occurence is ``0``.
        """
        change = True
        while self.content is not None and change:
            change = self.content.simplify()
            if self.content.maximum == 0 or len(self.content.target) == 0:
                self.content = None
            elif (len(self.content.target) == 1 and
                  isinstance(self.content.target.children[0].target, Group)):
                occur = self.content.target.children.pop()
                occur.minimum *= self.content.minimum
                if ( occur.maximum == INFINITY
                     or self.content.maximum == INFINITY):
                    occur.maximum = INFINITY
                else:
                    occur.maximum *= self.content.maximum
                self.content = occur
                change = True
        for att_uid, occur in self.attributes.items():
            if occur.maximum == 0:
                del(self.attributes[att_uid])

    def sort(self):
        """
        Calls the ``sort`` method on the group that describes the content
        of the element (if any) and its sub-groups. Also sorts the parents
        list.
        """
        self.parent_elts.sort(key=operator.attrgetter("qname"))
        if self.content is not None:
            self.content.target.sort()

    def finalize(self):
        self.simplify()
        self.compute_content_type()
        self.sort()
        # Builds the attributes and sub-elements dictionaries for the
        # analysis of the uses inside a library of XML documents
        self.actual_uses = {}
        self.attr_uses = {}
        for occ in self.iter_attrs():
            self.attr_uses[occ.target] = {}
        self.sub_elt_uses = {}
        if self.content is not None:
            for sub_elt in self.content.target.list_sub_elts(True):
                self.sub_elt_uses[sub_elt] = {}

    def reset_uses(self):
        self.actual_uses = {}
        for attr in self.attr_uses:
            self.attr_uses[attr] = {}
        for sub_elt in self.sub_elt_uses:
            self.sub_elt_uses[sub_elt] = {}

    def find_attributes(self, qname):
        """
        Finds and returns all the attributes with this qname.
        """
        occ = self.attributes.get(qname)
        if occ is not None:
            return [occ.target]
        else:
            return []

    def find_children(self, qname):
        """
        Finds and returns all the children with this qname.

        Only works after element has been finalized
        """
        return [sub_elt for sub_elt in self.sub_elt_uses
                if sub_elt.qname == qname]

    def find_parents(self, qname):
        """
        Finds and returns all the parents with this qname.
        """
        return [par_elt for par_elt in self.parent_elts
                if par_elt.qname == qname]

    def __repr__(self):
        return u"Element <{0}>".format(self.qname)


class XMLAttribute(XMLNode):
    """
    Class describing an XML attribute. Attributes occur inside elements.

    .. automethod:: __init__
    """
    def __init__(self, qname):
        """
        Initializes a new XML attribute.

        :param qname:
            Qualified name of the XML attribute (namespace + local name). The
            qualified name is expected to be in the ``{namespace}local-name``
            form.
        :type qname:
            :class:`unicode`
        """
        super(XMLAttribute, self).__init__(qname)

    def add_parent_elt(self, parent):
        """
        Adds a parent element for this attribute.

        :param parent:
            Parent element of this attribute (element that contains the
            attribute).
        :type parent:
            :class:`~pyxst.xml_struct.graph_nodes.XMLElement`
        """
        if parent not in self.parent_elts:
            if len(self.parent_elts) > 0:
                raise Exception("Attribute %s can't have two different parents"
                                % self.qname)
            self.parent_elts.append(parent)

    def __eq__(self, other):
        if not isinstance(other, XMLAttribute):
            return NotImplemented
        return (
            self.namespace == other.namespace
            and self.textual_content_type == other.textual_content_type
            and self.textual_content_is_list == other.textual_content_is_list
            and self.textual_content_values == other.textual_content_values)

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return u"Attribute @{0}".format(self.qname)



"""
Module defining a specific exception for the errors occuring while analyzing
an XML Schema.

.. autoclass:: XsdError
   :members:
   :undoc-members:
   :show-inheritance: 
"""

class XsdError(Exception):
    """
    Exception raised when finding an error while analyzing an XML Schema.
    """
    pass

